'''
TornadoServer
*************

Represents an endpoint to communicate with using Tornado asynchronous HTTP
client::

    from tornado import gen, ioloop

    @gen.coroutine
    def test():
        proxy = TornadoServer('http://example.com/api')
        result = yield proxy.some_method(42)
        print(result)

    ioloop.IOLoop.instance().run_sync(test)

'''

from tornado.httpclient import AsyncHTTPClient, HTTPError
from tornado.concurrent import Future

from .server import Server


class TornadoServer(Server):
    '''
    :param endpoint: The server address.
    :param async_http_client_class: Tornado asynchronous HTTP client class.
    :param kwargs: Keyword arguments to pass to the client initialiser.
    '''

    _default_headers = {
        'Content-Type' : 'application/json',
        'Accept'       : 'application/json'
    }


    def __init__(self, endpoint, async_http_client_class=AsyncHTTPClient, \
            **kwargs):
        super(TornadoServer, self).__init__(endpoint)

        self.http_client = async_http_client_class(**kwargs)

    def _process_response(self, response_future):
        future = Future()

        def process_response_callback(response_future):
            try:
                response = response_future.result()
                result = super(TornadoServer, self)._process_response(
                    response.body.decode())
            except Exception as ex: # pylint: disable=broad-except
                future.set_exception(ex)
            else:
                future.set_result(result)

        response_future.add_done_callback(process_response_callback)

        return future

    def _log_response_callback(self, future):
        ex = future.exception()
        if ex:
            if isinstance(ex, HTTPError):
                body = ex.response.body.decode() \
                    if ex.response and ex.response.body else None
                self._log_response(body, {
                    'http_code'    : ex.code,
                    'http_reason'  : ex.message,
                })
        else:
            response = future.result()
            self._log_response(response.body.decode(), {
                'http_code'    : response.code,
                'http_reason'  : response.reason,
                'http_headers' : response.headers
            })

    def _send_message(self, request, **kwargs):
        """Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :param kwargs: Keyword arguments to the Tornado request.
        :return: The response (a string for requests, None for notifications).
        """

        self._log_request(request)

        headers = dict(self._default_headers)
        headers.update(kwargs.get('headers', {}))

        future = self.http_client.fetch(self.endpoint, method='POST', \
            body=request, headers=headers, **kwargs)

        future.add_done_callback(self._log_response_callback)

        return future



"""setup.py"""

from codecs import open as codecs_open
from setuptools import setup

with codecs_open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()
with codecs_open('HISTORY.rst', 'r', 'utf-8') as f:
    history = f.read()

setup(
    name='jsonrpcclient',
    version='2.2.1',
    description='Send JSON-RPC requests',
    long_description=readme + '\n\n' + history,
    author='Beau Barker',
    author_email='beauinmelbourne@gmail.com',
    url='https://jsonrpcclient.readthedocs.io/',
    license='MIT',
    packages=['jsonrpcclient'],
    package_data={'jsonrpcclient': ['response-schema.json']},
    include_package_data=True,
    install_requires=['future', 'jsonschema'],
    extras_require={
        'requests' : ['requests'],
        'zmq'      : ['pyzmq'],
        'tornado'  : ['tornado']
    },
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



"""An HTTP server to communicate with, for example::

    HTTPServer('http://example.com/api').request('go')

Uses the Requests library.
http://docs.python-requests.org/en/master/
"""

from requests import Request, Session

from jsonrpcclient.server import Server


class HTTPServer(Server):
    """Defines an HTTP server"""

    # The default HTTP header
    __DEFAULT_HTTP_HEADERS__ = {
        'Content-Type': 'application/json', 'Accept': 'application/json'}

    def __init__(self, endpoint):
        """
        :param endpoint: The server address.
        :param kwargs: HTTP headers and other options passed on to the requests
                       module.
        """
        super(HTTPServer, self).__init__(endpoint)
        self.session = Session()
        self.session.headers.update(self.__DEFAULT_HTTP_HEADERS__)
        self.last_request = None
        self.last_response = None

    def _send_message(
            self, request, headers=None, files=None, params=None, auth=None,
            cookies=None, **kwargs):
        """Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :return: The JSON-RPC response.
        :rtype: A string for requests, None for notifications.
        :raise requests.exceptions.RequestException:
            Raised by the requests module in the event of a communications
            error.
        """
        # Prepare the request
        req = Request(
            method='POST', url=self.endpoint, data=request, headers=headers,
            files=files, params=params, auth=auth, cookies=cookies)
        prepped = self.session.prepare_request(req)
        self.last_request = prepped
        # Log the request
        self._log_request(request, {'http_headers': prepped.headers})
        # Send the message
        response = self.session.send(prepped, **kwargs)
        # Log the response
        self._log_response(response.text, {'http_code': response.status_code, \
            'http_reason': response.reason, 'http_headers': response.headers})
        self.last_response = response
        return response.text
